from yta_editor_nodes.abstract import _ProcessorGPUAndCPU
from yta_editor_nodes.utils import get_input_size
from typing import Union
from abc import abstractmethod, ABC


class _NodeProcessor(_ProcessorGPUAndCPU, ABC):
    """
    *For internal use only*

    *Singleton class*

    This class must be inherited by the specific
    implementation of some effect that will be done by
    CPU or GPU (at least one of the options)

    A simple processor node that is capable of
    processing inputs and obtain a single output by
    using the GPU or the CPU.

    This type of node is for the effects and 
    transitions.

    Mandatory parameters
    - `base_input`

    Optional parameters
    - `None`
    """

    mandatory_inputs = ['base_input']
    optional_inputs = []

    def __init__(
        self,
        opengl_context: Union['moderngl.Context', None],
        **kwargs
    ):
        """
        The `opengl_context` will be passed to the GPU node
        if needed (and existing).
        """
        node_complex_cpu, node_complex_gpu = self._instantiate_cpu_and_gpu_processors(
            opengl_context = opengl_context,
            **kwargs
        )

        super().__init__(
            processor_cpu = node_complex_cpu,
            processor_gpu = node_complex_gpu,
            opengl_context = opengl_context
        )

    @abstractmethod
    def _instantiate_cpu_and_gpu_processors(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        """
        *For internal use only*

        Instantiate the CPU and GPU procesors and return 
        them in that order.

        This method must be implemented in each class.
        """
        pass

    def _process_common(
        self,
        output_size: Union[tuple[int, int], None] = None,
        do_use_gpu: bool = True,
        **kwargs
    ):
        processor = self._get_processor(
            do_use_gpu = do_use_gpu
        )

        # The inputs must be transformed before reaching
        # this point

        # The **kwargs here will include inputs and other
        # parameters defined in the specific children

        return processor.process(
            output_size = output_size,
            **kwargs
        )
    
    def process(
        self,
        inputs: dict[str, Union['np.ndarray', 'moderngl.Texture']],
        output_size: Union[tuple[int, int], None] = None,
        do_use_gpu: bool = True,
        # This is to accept 't' even when needed
        **kwargs
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the provided `input` with GPU or CPU 
        according to the internal flag.
        """
        self._validate_inputs(inputs)

        base_input = inputs['base_input']

        output_size = (
            get_input_size(base_input)
            if output_size is None else
            output_size
        )

        processor = self._get_processor(
            do_use_gpu = do_use_gpu
        )

        return processor.process(
            # TODO: What do we do with this (?)
            inputs = {
                'base_input': self._prepare_input(
                    input = base_input,
                    do_use_gpu = do_use_gpu
                ),
            },
            output_size = output_size,
            **kwargs
        )


from yta_editor_nodes.complex.base import _NodeComplex
from yta_editor_nodes.utils import get_input_size
from typing import Union


class DisplayOverAtNodeComplex(_NodeComplex):
    """
    The overlay input is positioned with the given position,
    rotation and size, and then put as an overlay of the
    also given base input.

    Information:
    - The scene size is `(1920, 1080)`, so provide the
    `position` parameter according to it, where it is
    representing the center of the texture.
    - The `rotation` is in degrees, where `rotation=90`
    means rotating 90 degrees to the right.
    - The `size` parameter must be provided according to
    the previously mentioned scene size `(1920, 1080)`.

    TODO: This has no inheritance, is special and we need
    to be able to identify it as a valid one.
    """

    mandatory_inputs = ['base_input', 'overlay_input']
    optional_inputs = []

    def _instantiate_cpu_and_gpu_processors(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        """
        *For internal use only*

        Instantiate the CPU and GPU blenders and return 
        them in that order.

        This method must be implemented by each class.
        """
        from yta_editor_nodes_cpu.complex.display_over_at import DisplayOverAtNodeComplexCPU
        from yta_editor_nodes_gpu.complex.display_over_at import DisplayOverAtNodeComplexGPU

        node_cpu = DisplayOverAtNodeComplexCPU()
        node_gpu = DisplayOverAtNodeComplexGPU(
            opengl_context = opengl_context
        )

        return node_cpu, node_gpu
    
    def process(
        self,
        inputs: Union['np.ndarray', 'moderngl.Texture'],
        output_size: Union[tuple[int, int], None] = None,
        do_use_gpu: bool = True,
        position: tuple[int, int] = (1920 / 2, 1080 / 2),
        size: tuple[int, int] = (1920 / 2, 1080 / 2),
        rotation: int = 0
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Put the `overlapy_input` over the `base_input` in
        the `position` given and with the `size` and
        `rotation` also provided.
        """
        base_input = inputs['base_input']
        overlay_input = inputs['overlay_input']

        output_size = (
            get_input_size(base_input)
            if output_size is None else
            output_size
        )

        processor = self._get_processor(
            do_use_gpu = do_use_gpu
        )

        return processor.process(
            inputs = {
                'base_input': self._prepare_input(
                    input = base_input,
                    do_use_gpu = do_use_gpu
                ),
                'overlay_input': self._prepare_input(
                    input = overlay_input,
                    do_use_gpu = do_use_gpu
                ),
            },
            output_size = output_size,
            position = position,
            size = size,
            rotation = rotation
        )


from yta_video_opengl.context import OpenGLContext
from yta_validation import PythonValidator
from yta_editor_utils.texture import TextureUtils
from yta_editor_nodes_common import _Node
from abc import abstractmethod
from typing import Union


class _ProcessorGPUAndCPU(_Node):
    """
    *For internal use only*

    *Abstract class*

    Abstract class to share the common behaviour of
    being able to handle a process with a GPU and/or
    a CPU processor, chosen by the user.
    """

    mandatory_inputs = []
    optional_inputs = []

    @property
    def is_gpu_available(
        self
    ) -> bool:
        """
        Boolean flag to indicate if the GPU is available
        or not, that means that the processor that uses
        GPU is set.
        """
        return self._processor_gpu is not None
    
    @property
    def is_cpu_available(
        self
    ) -> bool:
        """
        Boolean flag to indicate if the CPU is available
        or not, that means that the processor that uses
        CPU is set.
        """
        return self._processor_cpu is not None
    
    def __init__(
        self,
        processor_cpu: Union['_ProcessorGPU', None] = None,
        processor_gpu: Union['_ProcessorCPU', None] = None,
        opengl_context: Union['moderngl.Context', None] = None,
    ):
        """
        The `processor_cpu` and `processor_gpu` have to be
        set by the developer when building the specific
        classes, but the `do_use_gpu` boolean flag will be
        set by the user when instantiating the class to
        choose between GPU and CPU.

        The `opengl_context` is needed to instantiate the
        GPU unit but also to transform the input into a
        `moderngl.Texture` if needed.
        """
        if (
            processor_cpu is None and
            processor_gpu is None
        ):
            raise Exception('No node processor provided. At least one node processor is needed.')

        self._processor_cpu: Union['_ProcessorCPU', None] = processor_cpu
        """
        *For internal use only*

        The transition processor that is able to do the
        processing by using the CPU. If it is None we cannot
        process it with CPU.
        """
        self._processor_gpu: Union['_ProcessorGPU', None] = processor_gpu
        """
        *For internal use only*

        The transition processor that is able to do the
        processing by using the GPU. If it is None we cannot
        process it with GPU.
        """
        self._opengl_context: Union['moderngl.Context', None] = (
            OpenGLContext().context
            if opengl_context is None else
            opengl_context
        )
        """
        *For internal use only*

        The context that we will use for the GPU node.
        """

    def _validate_inputs(
        self,
        inputs: dict[str, Union['np.ndarray', 'moderngl.Texture']]
    ):
        return super()._validate_inputs(inputs, ['ndarray', 'Texture'])

    def _get_processor(
        self,
        do_use_gpu: bool
    ) -> Union['ProcessorGPU', 'ProcessorCPU']:
        """
        *For internal use only*

        Get the processor according to the `do_use_gpu`
        parameter provided and if the one requested is
        available or not.
        """
        return (
            (
                # Prefer GPU if available
                self._processor_gpu or
                self._processor_cpu
            ) if do_use_gpu else (
                # Prefer CPU if available
                self._processor_cpu or
                self._processor_gpu
            )
        )
    
    # TODO: Add a new method that accepts 'inputs' and
    # handle them as mandatory and optional and prepare
    # them to be used
    
    # TODO: Maybe move to a more general thing (?)
    def _prepare_input(
        self,
        input: Union['moderngl.Texture', 'np.ndarray'],
        do_use_gpu: bool
    ) -> Union['moderngl.Texture', 'np.ndarray']:
        """
        *For internal use only*

        Prepare the `input` provided based on the `do_use_gpu`
        parameter given to be used with the processer,
        according to the availability.

        This method will transform a `texture` into a `numpy`
        array if they are asking for CPU and it is available,
        or a `numpy` into a `texture` if they are asking for
        GPU and it is available.
        """
        # 1. Get the backend we need
        if do_use_gpu and self.is_gpu_available:
            backend = 'gpu'
        elif not do_use_gpu and self.is_cpu_available:
            backend = 'cpu'
        elif self.is_gpu_available:
            backend = 'gpu'
        elif self.is_cpu_available:
            backend = 'cpu'
        else:
            raise RuntimeError("Node has neither CPU nor GPU implementation")

        return (
            # Is texture but ask CPU and available
            TextureUtils.texture_to_numpy(
                texture = input,
                do_include_alpha = True
            )
            if (
                backend == 'cpu' and
                not PythonValidator.is_numpy_array(input)
            ) else
            # Is numpy but ask GPU and available
            TextureUtils.numpy_to_texture(
                input = input,
                opengl_context = self._opengl_context
            )
            if (
                backend == 'gpu' and
                PythonValidator.is_numpy_array(input)
            ) else
            input
        )

    @abstractmethod
    def process(
        self,
        inputs: dict[str, Union['np.ndarray', 'moderngl.Texture']],
        do_use_gpu: bool = True
    ) -> Union['moderngl.Texture', 'np.ndarray']:
        """
        Process the provided `input` with GPU or CPU 
        according to the internal flag.
        """
        pass


from yta_editor_nodes.processor.base import _NodeProcessor
from yta_editor_nodes.utils import get_input_size
from typing import Union


class BlackAndWhiteNodeProcessor(_NodeProcessor):
    """
    The node to modify the input provided and set
    it as black and white by using CPU or GPU.

    Mandatory parameters
    - `base_input`

    Optional parameters
    - `None`
    """

    mandatory_inputs = ['base_input']
    optional_inputs = []

    def __init__(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        super().__init__(
            opengl_context = opengl_context
        )

    def _instantiate_cpu_and_gpu_processors(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        """
        *For internal use only*

        Instantiate the CPU and GPU procesors and return 
        them in that order.

        This method must be implemented by each class.
        """
        from yta_editor_nodes_cpu.processor.black_and_white import BlackAndWhiteNodeProcessorCPU
        from yta_editor_nodes_gpu.processor.black_and_white import BlackAndWhiteNodeProcessorGPU

        node_cpu = BlackAndWhiteNodeProcessorCPU()
        node_gpu = BlackAndWhiteNodeProcessorGPU(
            opengl_context = opengl_context
        )

        return node_cpu, node_gpu

    def process(
        self,
        inputs: dict[str, Union['np.ndarray', 'moderngl.Texture']],
        output_size: Union[tuple[int, int], None] = None,
        do_use_gpu: bool = True,
        # This is to accept 't' even when unneeded
        **kwargs
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the provided `input` with GPU or CPU 
        according to the internal flag.
        """
        return super().process(
            inputs = inputs,
            output_size = output_size,
            do_use_gpu = do_use_gpu
        )


from yta_editor_nodes.blender.base import _Blender
from typing import Union


class AddBlender(_Blender):
    """
    The most common blender used in video edition.

    This blender will increase the brightness by
    combining the colors of the base and the overlay
    inputs, using the overlay as much as the 
    `strength` parameter is indicating.
    """

    mandatory_inputs = ['base_input', 'overlay_input']
    optional_inputs = []

    def __init__(
        self,
        opengl_context: Union['moderngl.Context', None],
    ):
        super().__init__(
            opengl_context = opengl_context
        )

    def _instantiate_cpu_and_gpu_blenders(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        """
        *For internal use only*

        Instantiate the CPU and GPU blenders and return 
        them in that order.

        This method must be implemented by each class.
        """
        from yta_editor_nodes_cpu.blender.add import AddBlenderCPU
        from yta_editor_nodes_gpu.blender.add import AddBlenderGPU

        node_cpu = AddBlenderCPU()
        node_gpu = AddBlenderGPU(
            opengl_context = opengl_context
        )

        return node_cpu, node_gpu

    def blend(
        self,
        inputs: dict[str, Union['np.ndarray', 'moderngl.Texture']],
        output_size: Union[tuple[int, int], None] = None,
        do_use_gpu: bool = True,
        mix_weight: float = 1.0,
        strength: float = 1.0
    # TODO: What about the output type (?)
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the provided 'base_input' and 'overlay_input'
        with GPU or CPU according to the internal flag.

        The `mix_weight` will determine how much of the result
        we keep against the base, where a `mix_weight == 1.0`
        means that we want the result at 100%, but a
        `mix_weight == 0.4` means a 40% of the result and a 60%
        of the base.
        """
        return self.process(
            inputs = inputs,
            output_size = output_size,
            do_use_gpu = do_use_gpu,
            mix_weight = mix_weight,
            strength = strength
        )
    
    # TODO: This method has to be reviewed in the general
    # class to be able to receive array parameters and
    # send them to the individual process as single parameter
    def blend_multiple_inputs(
        self,
        inputs: list[Union['np.ndarray', 'moderngl.Texture']],
        output_size: Union[tuple[int, int], None] = None,
        do_use_gpu: bool = True,
        mix_weights: Union[list[float], float] = 1.0,
        strengths: Union[list[float], float] = 1.0,
    # TODO: What about the output type (?)
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the provided 'inputs' with GPU or CPU
        according to the internal flag.
        """
        return self.blend_multiple_inputs(
            inputs = inputs,
            output_size = output_size,
            do_use_gpu = do_use_gpu,
            mix_weights = mix_weights,
            strengths = strengths,
        )


from yta_editor_nodes.compositor.base import _NodeCompositor
from yta_editor_nodes.utils import get_input_size
from typing import Union


class DisplacementWithRotationNodeCompositor(_NodeCompositor):
    """
    The frame, but moving and rotating over other frame.

    This node uses these inputs:
    - `base_input`
    - `overlay_input`
    """

    mandatory_inputs = ['overlay_input']
    optional_inputs = ['base_input']

    def _instantiate_cpu_and_gpu_processors(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        """
        *For internal use only*

        Instantiate the CPU and GPU procesors and return 
        them in that order.

        This method must be implemented by each class.
        """
        from yta_editor_nodes_cpu.compositor.displacement_with_rotation import DisplacementWithRotationNodeCompositorCPU
        from yta_editor_nodes_gpu.compositor.displacement_with_rotation import DisplacementWithRotationNodeCompositorGPU

        node_cpu = DisplacementWithRotationNodeCompositorCPU()
        node_gpu = DisplacementWithRotationNodeCompositorGPU(
            opengl_context = opengl_context
        )

        return node_cpu, node_gpu
    
    def process(
        self,
        inputs: dict[str, Union['np.ndarray', 'moderngl.Texture']],
        output_size: Union[tuple[int, int], None] = None,
        # output_width: Union[int, None] = None,
        # output_height: Union[int, None] = None,
        do_use_gpu: bool = True,
        x: int = 1920 / 2 * 1.5,
        y: int = 1080 / 2 * 1.5,
        width: int = (1920 / 2 * 1.5),
        height: int = (1080 / 2 * 1.5),
        rotation: int = 45,
        **kwargs
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the provided 'input' with GPU or CPU 
        according to the internal flag.
        """
        # output_size = (output_width, output_height)
        base_input = inputs.get('base_input', None)
        overlay_input = inputs['overlay_input']

        input_for_size = (
            base_input
            if base_input is not None else
            overlay_input
        )

        output_size = (
            get_input_size(input_for_size)
            if output_size is None else
            output_size
        )

        processor = self._get_processor(
            do_use_gpu = do_use_gpu
        )

        base_input = (
            self._prepare_input(
                input = base_input,
                do_use_gpu = do_use_gpu
            )
            if base_input is not None else
            None
        )

        # Built dynamic parameters
        position = (x, y)
        size = (width, height)

        return processor.process(
            inputs = {
                'base_input': base_input,
                'overlay_input': self._prepare_input(
                    input = overlay_input,
                    do_use_gpu = do_use_gpu
                ),
            },
            output_size = output_size,
            position = position,
            size = size,
            rotation = rotation
        )


from yta_editor_nodes.processor.video.transitions.base import _TransitionProcessor
from typing import Union


class CrossfadeTransitionProcessor(_TransitionProcessor):
    """
    A transition between the frames of 2 videos,
    transforming the first one into the second one.

    Class that is capable of doing some processing on
    an input by using the GPU or the CPU.
    """

    mandatory_inputs = ['first_input', 'second_input']
    optional_inputs = []

    def _instantiate_cpu_and_gpu_processors(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        """
        *For internal use only*

        Instantiate the CPU and GPU processors and return 
        them in that order.
        """
        from yta_editor_nodes_gpu.processor.video.transitions.crossfade import CrossfadeTransitionNodeProcessorGPU

        node_cpu = None
        node_gpu = CrossfadeTransitionNodeProcessorGPU(
            opengl_context = opengl_context
        )

        return node_cpu, node_gpu


"""
TODO: Maybe this is not the place to be in.
"""
from yta_validation.parameter import ParameterValidator
from yta_validation import PythonValidator
from typing import Union


class _TimedNode:
    """
    *For internal use only*

    Class to represent a Node wrapper to
    be able to specify the time range in
    which we want the node to be applied.

    If the 't' time moment is not inside
    this range, the frame will be returned
    as it is, with no change.

    A 't' time moment inside the range has
    this condition:
    - `start <= t < end`

    We are not including the end because
    the next TimedNode could start on that
    specific value, and remember that the
    first time moment is 0.

    This is the class that has to be applied
    when working with videos and not a Node
    directly.

    The 'start' and 'end' values by default
    are 0 and the end of the media to which
    we will apply the modification (that is
    defined here by None).
    """

    @property
    def copy(
        self
    ) -> '_TimedNode':
        """
        Get a copy of this instance.
        """
        return self.__class__(
            node = self.node,
            start = self.start,
            end = self.end
        )

    @property
    def is_audio_node(
        self
    ) -> bool:
        """
        Flag to indicate if the node (or effect)
        is an audio effect or not.
        """
        # TODO: Is this checking not only the
        # first level (?)
        return PythonValidator.is_subclass_of(self.node, '_AudioNodeProcessor')
    
    @property
    def is_video_node(
        self
    ) -> bool:
        """
        Flag to indicate if the node (or effect)
        is a video effect or not.
        """
        # TODO: Is this checking not only the
        # first level (?)
        return PythonValidator.is_subclass_of(self.node, '_VideoNodeProcessor')

    def __init__(
        self,
        node: Union['_VideoNodeProcessor', '_AudioNodeProcessor'],
        start: Union[int, float, 'Fraction'] = 0.0,
        end: Union[int, float, 'Fraction', None] = None
    ):
        ParameterValidator.validate_mandatory_positive_number('start', start, do_include_zero = True)
        ParameterValidator.validate_positive_number('end', end, do_include_zero = False)

        if (
            end is not None and
            end < start
        ):
            raise Exception('The "end" parameter provided must be greater or equal to the "start" parameter.')

        self.node: Union['_VideoNodeProcessor', '_AudioNodeProcessor'] = node
        """
        The node we are wrapping and we want to
        apply as a modification of the frame in
        which we are in a 't' time moment.
        """
        self.start: float = start
        """
        The 't' time moment in which the Node must
        start being applied (including it).
        """
        # TODO: The 'end' we receive here could be
        # greater than the actual end of the media
        # in which the TimedNode will be applied,
        # but it seems to be working correctly as
        # we will never receive a 't' from that
        # media that is out of its own bounds...
        self.end: Union[float, None] = end
        """
        The 't' time moment in which the Node must
        stop being applied (excluding it).
        """

    def _get_t(
        self,
        t: Union[int, float, 'Fraction']
    ) -> float:
        """
        Obtain the 't' time moment relative to the
        effect duration.

        Imagine `start=3` and `end=5`, and we receive 
        a `t=4`. It is inside the range, so we have
        to apply the effect, but as the effect
        lasts from second 3 to second 5 (`duration=2`),
        the `t=4` is actually a `t=1` for the effect
        because it is the time elapsed since the
        effect started being applied, that was on the 
        second 3.

        The formula:
        - `t - self.start`
        """
        return t - self.start

    def is_within_time(
        self,
        t: float
    ) -> bool:
        """
        Flag to indicate if the 't' time moment provided
        is in the range of this TimedNode instance, 
        which means between the 'start' and the 'end'.

        The formula:
        - `start <= t < end`
        """
        return (
            self.start <= t < self.end
            if self.end is not None else
            self.start <= t
        )

    def process(
        self,
        frame: Union['moderngl.Texture', 'np.ndarray'],
        t: float
        # TODO: Maybe we need 'fps' and 'number_of_frames'
        # to calculate progressions or similar...
    ) -> Union['moderngl.Texture', 'np.ndarray']:
        """
        Process the frame if the provided 't' time
        moment is in the range of this TimedNode
        instance.
        """
        return (
            self.node.process(frame, self._get_t(t))
            if self.is_within_time(t) else
            frame
        )
    
class TimedVideoNode(_TimedNode):
    """
    Class to represent a video node wrapper
    to be able to specify the time range in
    which we want the node to be applied.

    If the 't' time moment is not inside
    this range, the frame will be returned
    as it is, with no change.

    A 't' time moment inside the range has
    this condition:
    - `start <= t < end`

    We are not including the end because
    the next TimedNode could start on that
    specific value, and remember that the
    first time moment is 0.

    This is the class that has to be applied
    when working with videos and not a Node
    directly.

    The 'start' and 'end' values by default
    are 0 and the end of the media to which
    we will apply the modification (that is
    defined here by None).
    """

    def __init__(
        self,
        node: '_VideoNodeProcessor',
        start: Union[int, float, 'Fraction'] = 0.0,
        end: Union[int, float, 'Fraction', None] = None
    ):
        # TODO: Is this working (?)
        ParameterValidator.validate_mandatory_subclass_of('node', node, '_VideoNodeProcessor')

        super().__init__(
            node = node,
            start = start,
            end = end
        )

class TimedAudioNode(_TimedNode):
    """
    Class to represent an audio node wrapper
    to be able to specify the time range in
    which we want the node to be applied.

    If the 't' time moment is not inside
    this range, the frame will be returned
    as it is, with no change.

    A 't' time moment inside the range has
    this condition:
    - `start <= t < end`

    We are not including the end because
    the next TimedNode could start on that
    specific value, and remember that the
    first time moment is 0.

    This is the class that has to be applied
    when working with videos and not a Node
    directly.

    The 'start' and 'end' values by default
    are 0 and the end of the media to which
    we will apply the modification (that is
    defined here by None).
    """

    def __init__(
        self,
        node: '_AudioNodeProcessor',
        start: Union[int, float, 'Fraction'] = 0.0,
        end: Union[int, float, 'Fraction', None] = None
    ):
        # TODO: Is this working (?)
        ParameterValidator.validate_mandatory_subclass_of('node', node, '_AudioNodeProcessor')

        super().__init__(
            node = node,
            start = start,
            end = end
        )


from yta_editor_nodes.timeline import TimelineNode
from yta_editor_nodes.timeline.serial import SerialTimelineNode
from yta_editor_time.specifications import TimeSpecification
from abc import ABC, abstractmethod
from typing import Union


class _ParallelTimelineNodeAbstract(TimelineNode, ABC):
    """
    The abstract class of the parallel node.

    This class is limited and needs at least 2 nodes
    to be valid.
    """

    def __init__(
        self,
        # TODO: Put the correct class
        # TODO: Maybe I need the 'node' in the TimelineNode class
        # TODO: Maybe I need a shortcut to 'is_gpu_available'...
        name: str,
        nodes: list[SerialTimelineNode]
    ):
        if len(nodes) < 2:
            raise Exception('A ParallelTimelineNode needs 2 or more nodes to be created.')
        
        self.nodes: list[SerialTimelineNode] = nodes
        """
        The list of nodes that will be executed in parallel
        to obtain the outputs that will be combined.
        """

        # Calculate the start and end based on nodes
        t_start = min(float(node.specification.start_t) for node in nodes)
        t_end = max(float(node.specification.end_t) for node in nodes)

        super().__init__(
            name = name,
            specification = TimeSpecification(
                start_t = t_start,
                end_t = t_end,
                duration = None
            )
        )

    def _process(
        self,
        inputs: dict[str, Union['np.ndarray', 'moderngl.Texture']],
        evaluation_context: 'EvaluationContext',
        **kwargs
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the `input` provided for the `t` time moment
        given and combine the outputs of the different
        internal nodes and return them as output.
        """
        active_nodes = [
            node for node in self.nodes
            if node.is_active_at(evaluation_context.t)
        ]

        # This shouldn't happen as we know that the parallel
        # node itself is not active, but just in case...
        if not active_nodes:
            # TODO: What do we do with the different inputs? We
            # should return only one from the dict...
            return inputs

        if len(active_nodes) == 1:
            return active_nodes[0].process(
                inputs = inputs,
                evaluation_context = evaluation_context,
                **kwargs
            )

        outputs = [
            node.process(
                inputs = inputs,
                evaluation_context = evaluation_context,
                **kwargs
            )
            for node in active_nodes
        ]

        return self._combine_outputs(outputs)
    
    @abstractmethod
    def _combine_outputs(
        self,
        outputs: list[Union['np.ndarray', 'moderngl.Texture']]
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Combine the different `outputs` into a single one.

        This class must be overwritten by every specific
        class implementation.
        """
        # TODO: This must be according to the type
        pass


from yta_editor_nodes.processor.base import _ProcessorGPUAndCPU
from yta_editor_nodes.utils import get_input_size
from typing import Union
from abc import abstractmethod


class _NodeCompositor(_ProcessorGPUAndCPU):
    """
    *For internal use only*

    This class must be inherited by the specific
    implementation of some node that will be positioning
    inputs, done by CPU or GPU (at least one of the
    options)

    A node specifically designed to build a scene by
    positioning inputs in different positions and 
    obtaining a single output by using GPU or CPU.
    """

    mandatory_inputs = ['base_input']
    optional_inputs = []

    def __init__(
        self,
        opengl_context: Union['moderngl.Context', None],
        **kwargs
    ):
        node_compositor_cpu, node_compositor_gpu = self._instantiate_cpu_and_gpu_processors(
            opengl_context = opengl_context,
            **kwargs
        )

        super().__init__(
            processor_cpu = node_compositor_cpu,
            processor_gpu = node_compositor_gpu,
            opengl_context = opengl_context
        )

    @abstractmethod
    def _instantiate_cpu_and_gpu_processors(
        self,
        opengl_context: Union['moderngl.Context', None],
        **kwargs
    ):
        """
        *For internal use only*

        Instantiate the CPU and GPU procesors and return 
        them in that order.

        This method must be implemented by each class.
        """
        pass

    # TODO: Maybe @abstractmethod cause' they are different
    def process(
        self,
        inputs: dict[str, Union['np.ndarray', 'moderngl.Texture']],
        output_size: Union[tuple[int, int], None] = None,
        do_use_gpu: bool = True,
        **kwargs
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the provided `input` with GPU or CPU 
        according to the `do_use_gpu` flag provided.
        """
        base_input = inputs['base_input']

        output_size = (
            get_input_size(base_input)
            if output_size is None else
            output_size
        )

        processor = self._get_processor(
            do_use_gpu = do_use_gpu
        )

        return processor.process(
            input = self._prepare_input(
                input = input,
                do_use_gpu = do_use_gpu
            ),
            **kwargs
        )


"""
The blenders, capable of mixing different inputs into
a single one by using different parameters and 
techniques.

All the classes here will have an instance of the
specific CPU and/or GPU class that is able to run the
code by using either CPU or GPU. The user can choose
between GPU and CPU and that option will be considered
(only if available).
"""
from yta_editor_nodes.blender.mix import MixBlender
from yta_editor_nodes.blender.alpha import AlphaBlender
from yta_editor_nodes.blender.add import AddBlender


__all__ = [
    'MixBlender',
    'AlphaBlender',
    'AddBlender'
]

"""
Note for the developer:

I leave this old way to instantiate dynamically
but less refactored, just in case. Please, remove
in the next commits if the new version is already
working.

TODO: Old way to import dynamically
blender_cpu = None
if is_cpu_available():
    from yta_editor_nodes_cpu.blender.add import AddBlenderCPU

    blender_cpu = AddBlenderCPU()

blender_gpu = None
if is_gpu_available():
    from yta_editor_nodes_gpu.blender import AddBlenderGPU

    blender_gpu = AddBlenderGPU(
        opengl_context = None,
        # TODO: Do not hardcode please...
        output_size = (1920, 1080),
    )
"""


from yta_editor_parameters.abstract import VideoEditorParameter
from yta_editor_time.specifications import FrameIndexSpecification, ProgressSpecification, TimeSpecification
from yta_editor_time.evaluation_context import EvaluationContext
from yta_validation.parameter import ParameterValidator
from abc import ABC, abstractmethod
from typing import Union


class TimelineNode(ABC):
    """
    *Abstract class*

    *This class has to be inherited by any class that is
    able to handle some input to obtain an output as a
    result*

    The abstract class of a TimelineNode, which is the
    entity that is able to process some input to return
    an output that can be sent to the next node, and able
    to connect (by storing the references) to the other
    nodes.
    """

    @property
    def has_output_node(
        self
    ) -> bool:
        """
        Boolean flag to indicate if this node has an
        output node or not.
        """
        return self.output_node is not None
    
    @property
    def has_input_node(
        self
    ) -> bool:
        """
        Boolean flag to indicate if this node has an
        input node or not.
        """
        return self.input_node is not None
    
    @property
    def duration(
        self
    ) -> float:
        """
        The duration of the timeline node, in seconds.

        The formula:
        - `self.t_end - self.t_start`
        """
        return self.t_end - self.t_start

    def __init__(
        self,
        # TODO: We need a reference to the Timeline to know fps
        name: str,
        # TODO: By now I'm using 'TimeSpecification' because
        # of how it was built before, but I should accept any
        # of them
        specification: TimeSpecification,
        parameters: dict[str, VideoEditorParameter] = {}
    ):
        ParameterValidator.validate_mandatory_string('name', name, do_accept_empty = False)
        # TODO: Validate 'parameters' as dict of 
        # 'str': VideoEditorParameter

        self.name: str = name
        """
        Just a simple name for the node.
        """
        self.is_enabled: bool = True
        """
        Boolean flag to indicate if the node is enabled or not,
        which means that it will not be working if not enabled.
        Basically, if it is on or off.
        """
        self.specification: TimeSpecification = specification
        """
        The specification of when we should apply the settings,
        that can be based on frame index, time or normalized
        progress.
        """
        self.parameters: dict[str, VideoEditorParameter] = parameters
        """
        A dict including the `key` of the parameter and
        its value to be calculated.
        """

        self.input_node: Union['TimelineNode', None] = None
        """
        The node that is connected to this one as an
        input, so the output of that previous node will
        be passed as the input for this one.
        """
        self.output_node: Union['TimelineNode', None] = None
        """
        This node is connected as the input of another
        one that is after this. The output of this one
        will be sent as the input for that one.
        """
        self._cached_output: Union['np.ndarray', 'moderngl.Texture', None] = None
        """
        The output, but cached, so when it is generated
        it is stored here.
        """

    # TODO: Should we transform this into a progress (?)
    def _get_t(
        self,
        t: Union[int, float, 'Fraction']
    ) -> float:
        """
        Obtain the `t` time moment relative to the
        effect duration.

        Imagine `start=3` and `end=5`, and we receive 
        a `t=4`. It is inside the range, so we have
        to apply the effect, but as the effect
        lasts from second 3 to second 5 (`duration=2`),
        the `t=4` is actually a `t=1` for the effect
        because it is the time elapsed since the
        effect started being applied, that was on the 
        second 3.

        The formula:
        - `t - self.specification.start_t`
        """
        return float(t) - float(self.specification.start_t)
    
    def enable(
        self
    ) -> 'TimelineNode':
        """
        Enable the node by setting the internal flag to
        `True`, which means that will be working.
        """
        self.is_enabled = True

        return self
    
    def disable(
        self
    ) -> 'Timeline':
        """
        Disable the node by setting the internal flag
        to `False`, which means that will be working not.
        """
        self.is_enabled = False

        return self
    
    def is_active_at(
        self,
        t_timeline: float
    ) -> bool:
        """
        Flag to indicate if the `t_timeline` time moment
        provided is in the range of this TimedNode instance, 
        which means between the `t_start` and the `t_end`,
        and if it is activated or not (on or off).

        The formula:
        - `t_start <= t_timeline < t_end`
        """
        # TODO: We need to resolve with the new methods
        # that are in 'yta_editor_time'
        from yta_editor_time.specifications.resolver import SpecificationResolver

        # TODO: We need the real values...
        fps = 60.0
        duration = 2.0

        frame_index_specification_resolved = SpecificationResolver.resolve(
            specification = self.specification,
            fps = fps,
            duration = duration
        )

        # By now we transform it into 't' to work
        # as we were used to

        t_start = int(frame_index_specification_resolved.start_frame) / fps
        t_end = int(frame_index_specification_resolved.end_frame) / fps

        return (
            self.is_enabled and
            t_start <= float(t_timeline) < t_end
        )

        return (
            self.is_enabled and
            self.t_start <= t_timeline < self.t_end
        )

    def connect_to(
        self,
        node: 'TimelineNode'
    ) -> 'TimelineNode':
        """
        Connect the `node` provided to this one as an
        output, and also this one as an input of the
        `node` provided.

        TODO: The connection has to be done with another
        class that inherits from this 'TimelineNode'.
        """
        self.output_node = node
        node.input_node = self

        return self

    def clear_cache(
        self
    ) -> 'TimelineNode':
        """
        Clear the cache of this node, but also for
        the nodes that are outputs of this one, as
        the results would change.
        """
        # TODO: I think this cache is not valid because
        # we are using a `t` and different inputs...
        self._cached_output = None

        self.output_node.clear_cache()

        return self
    
    # TODO: Review this and with the context
    def _evaluate_parameters(
        self,
        evaluation_context: 'EvaluationContext'
    ) -> dict[str, any]:
        """
        *For internal use only*

        Evaluate all the parameters for the given `t` time
        moment and obtain the values for that moment.
        """
        parameters_values = {}

        # t_local_normalized = self._get_t(t_timeline) / self.duration

        # TODO: Create the context as we do in 'yta_editor'
        # from yta_editor_parameters.context.utils import get_evaluation_context

        # evaluation_context = get_evaluation_context(
        #     t_track = t_timeline,
        #     t_start_item = self.t_start,
        #     t_end_item = self.t_end,
        #     # TODO: Need reference to Timeline
        #     # fps = self._timeline.fps,
        #     # track_duration = self._timeline.duration
        # )

        for key, parameter in self.parameters.items():
            parameters_values[key] = parameter.evaluate(
                evaluation_context = evaluation_context
            )

        return parameters_values
    
    def process(
        self,
        inputs: dict[str, Union['np.ndarray', 'moderngl.Texture']],
        evaluation_context: 'EvaluationContext',
        # TODO: I think this has to be gone
        **kwargs
    ) -> Union['moderngl.Texture', 'np.ndarray']:
        """
        Process the `inputs` by using the `t_timeline` time
        moment provided, that could be necessary for the
        node if it is a dynamic one that depends on this
        value.
        """
        # TODO: Maybe a copy of inputs (?)

        """
        TODO: This condition is evalulated in the
        TimelineGraph this node belongs to, but we need
        the `t` for some dynamic NodeProcessors
        """
        # if not self.is_active_at(t_timeline):
        #     return result

        # TODO: With simple nodes we don't need the 't' and
        # this is failing if we don't put the **kwargs in
        # our CPU or GPU processors
        # This will be ignored if no needed as we use **kwargs
        parameters_values = self._evaluate_parameters(
            evaluation_context = evaluation_context
        )

        # This should be 'float32' or 'Texture'
        return self._process(
            inputs = inputs,
            evaluation_context = evaluation_context,
            **parameters_values
            # **kwargs
        )

    @abstractmethod
    def _process(
        self,
        inputs: dict[str, Union['np.ndarray', 'moderngl.Texture']],
        evaluation_context: 'EvaluationContext',
        **kwargs
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the `inputs` and obtain a result as the
        output.

        This method must be overwritten for every special
        class implementation.
        """
        pass


"""
Our awesome editor module in which we have
all the classes that interact with it and
make it possible.

This is the nodes module, in which we have
all the classes that make the concept work.
"""
from yta_editor_nodes.timeline.utils import is_edition_node
from yta_editor_nodes.timeline.abstract import TimelineNode
from yta_editor_nodes.timeline.serial import SerialTimelineNode
from yta_editor_nodes.timeline.parallel import ParallelTimelineNode
from typing import Union


# TODO: This below is to test the timeline and graphs
class TimelineGraph:
    """
    Graph of nodes (serie and parallel) to process
    the input, handle it and generate an output.

    The graph includes a starting (root) node that
    will begin the process, and maybe other nodes
    connected to it as output nodes.

    The nodes will be interconnected, having a
    `t_start` and `t_end`. The sequence of nodes
    determines the order, and the time range if
    they should modify the input or not.
    """

    def __init__(
        self,
        # TODO: Know 'timeline' associated to know fps (?)
        root_node: Union[SerialTimelineNode, ParallelTimelineNode]
    ):
        self.root_node: Union[SerialTimelineNode, ParallelTimelineNode] = root_node
        """
        The root node in which everything starts.
        """
        self._last_node: Union[SerialTimelineNode, ParallelTimelineNode] = root_node
        """
        *For internal use only*

        The last node we have and the one that will return
        the output.
        """

    def _connect_node_to_last(
        self,
        node: Union[SerialTimelineNode, ParallelTimelineNode]
    ) -> 'TimelineGraph':
        """
        *For internal use only*

        Connect the `node` provided to the last node of this
        timeline, and set this new node as the last one.
        """
        self._last_node.connect_to(node)
        self._last_node = node

        return self

    def add_node(
        self,
        node: Union[SerialTimelineNode, ParallelTimelineNode]
    ) -> 'TimelineGraph':
        """
        Add the `node` provided to the nodes list, connecting
        it (in serie mode) to the last one.
        """
        # TODO: Validate node (?)
        return self._connect_node_to_last(node)

    def add_parallel_node(
        self,
        nodes: list[SerialTimelineNode]
    ):
        """
        Create a `ParallelTimelineNode` with the `nodes` provided
        as parameter and connect it (in serie mode) to the last
        one.
        """
        parallel_node = ParallelTimelineNode(
            name = 'invented',
            # TODO: Validate nodes (?)
            nodes = nodes
        )
        
        return self._connect_node_to_last(parallel_node)

    # TODO: Maybe rename to 'process' (?)
    def render(
        self,
        inputs: dict[str, Union['np.ndarray', 'moderngl.Texture']],
        t: float
    ):
        """
        Render the provided `inputs` at the given `t` time
        moment. The `inputs` must be the frame that belongs
        to that time moment.

        The different nodes of this instance will be applied
        in order and only if they are enabled and active in
        the `t` time moment provided.
        """
        # TODO: Maybe a '.copy()' instead (?)
        # TODO: This is not ok, it should be this but maybe
        # the first key that is available
        output = inputs['base_input']
        node = self.root_node

        from yta_editor_time.evaluation_context import EvaluationContext

        # TODO: Where do I get this from (?)
        fps = 60.0
        total_frames = 120

        evaluation_context = EvaluationContext(
            fps = fps,
            total_frames = total_frames,
            frame_index = int(t * fps)
        )

        while node is not None:
            if node.is_active_at(
                t_timeline = t
            ):
                output = node.process(
                    inputs = inputs,
                    evaluation_context = evaluation_context
                )

            node = node.output_node

        return output
    
__all__ = [
    'TimelineGraph',
    'SerialTimelineNode',
    'ParallelTimelineNode'
]
    

"""
Note for the developer:

A guide to the different types of nodes we will have
when imitating DaVinci Resolve.

TimelineNode (abstracto)
├── ProcessorNode (procesa un solo flujo)
│   ├── EffectNode (filtros, LUTs, shaders…)
│   └── TransitionNode (mezcla entre dos entradas)
│
├── CompositeNode (combina múltiples flujos)
│   ├── SerialTimelineNode (uno tras otro)
│   ├── ParallelTimelineNode (ramas simultáneas)
│   └── LayerNode (capas con blending o máscaras)
│
└── GroupNode (contiene una mini subred de nodos)

"""



from yta_editor_nodes.processor.video.transitions.base import _TransitionProcessor
from yta_editor_nodes.utils import get_input_size
from typing import Union


class AlphaPediaMaskTransitionProcessor(_TransitionProcessor):
    """
    A transition made by using a custom mask to
    join the 2 videos. This mask is specifically
    obtained from the AlphaPediaYT channel in which
    we upload specific masking videos.

    Both videos will be placed occupying the whole
    scene, just overlapping by using the transition
    video mask, but not moving the frame through 
    the screen like other classes do (like the
    FallingBars).

    Class that is capable of doing some processing on
    an input by using the GPU or the CPU.
    """

    mandatory_inputs = ['first_input', 'second_input', 'mask_input']
    optional_inputs = []

    def _instantiate_cpu_and_gpu_processors(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        """
        *For internal use only*

        Instantiate the CPU and GPU processors and return 
        them in that order.
        """
        from yta_editor_nodes_gpu.processor.video.transitions.alphapedia_mask import AlphaPediaMaskTransitionNodeProcessorGPU

        node_cpu = None
        node_gpu = AlphaPediaMaskTransitionNodeProcessorGPU(
            opengl_context = opengl_context,
        )

        return node_cpu, node_gpu
    
    def process(
        self,
        inputs: dict[str, Union['moderngl.Texture', 'np.ndarray']],
        progress: float,
        output_size: Union[tuple[int, int], None],
        do_use_gpu: bool = True
    ) -> Union['moderngl.Texture', 'np.ndarray']:
        """
        Process the transition between the given `first_input`
        and `second_input`, with GPU or CPU according to the
        internal flag.
        """
        self._validate_inputs(inputs)
        
        first_input = inputs['first_input']
        second_input = inputs['second_input']
        mask_input = inputs['mask_input']

        output_size = (
            get_input_size(first_input)
            if output_size is None else
            output_size
        )

        processor = self._get_processor(
            do_use_gpu = do_use_gpu
        )

        return processor.process(
            inputs = {
                'first_input': self._prepare_input(
                    input = first_input,
                    do_use_gpu = do_use_gpu
                ),
                'second_input': self._prepare_input(
                    input = second_input,
                    do_use_gpu = do_use_gpu
                ),
                'mask_input': self._prepare_input(
                    input = mask_input,
                    do_use_gpu = do_use_gpu
                )
            },
            progress = progress
        )


from yta_editor_nodes.timeline import TimelineNode
from yta_editor_parameters.abstract import VideoEditorParameter
from yta_editor_parameters.utils import parse_parameters
from yta_editor_time.specifications import FrameIndexSpecification, ProgressSpecification, TimeSpecification
from typing import Union


class SerialTimelineNode(TimelineNode):
    """
    A node that is executed from a specific input and
    generates a single output that is sent to the next
    node.

    Check the `.process` method definition of the node
    you add to see the parameters you need to execute
    it and pass them through the `parameters`
    parameter.
    """

    def __init__(
        self,
        # TODO: Put the correct class
        # TODO: Maybe I need the 'node' in the TimelineNode class
        # TODO: Maybe I need a shortcut to 'is_gpu_available'...
        node: 'NodeProcessor',
        name: str,
        specification: Union[FrameIndexSpecification, ProgressSpecification, TimeSpecification],
        parameters: dict[str, VideoEditorParameter] = {}
    ):
        self.node: 'NodeProcessor' = node
        """
        The node to execute and to obtain the output from.
        """
        # Validate the 'parameters' by accepting
        # basic and non iterable types. This will raise
        # exception if not valid
        parameters = parse_parameters(parameters)

        super().__init__(
            name = name,
            specification = specification,
            parameters = parameters
        )

    def _process(
        self,
        inputs: dict[str, Union['np.ndarray', 'moderngl.Texture']],
        evaluation_context: 'EvaluationContext',
        **kwargs
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the input and obtain a result as the output.
        """
        t_local = self._get_t(t = evaluation_context.t)

        # Add 't' as dynamic because some NodeProcessor
        # don't actually expect it as parameter
        kwargs['t'] = t_local

        return self.node.process(
            inputs = inputs,
            **kwargs
        )


from yta_editor_nodes.blender.base import _Blender
from typing import Union


class AlphaBlender(_Blender):
    """
    The most common blender used in video edition.

    This blender will use the alpha channel of the 
    overlay input, multiplied by the `blend_strength`
    parameter provided, to use it as the mixer factor
    between the base and the overlay inputs.
    """

    mandatory_inputs = ['base_input', 'overlay_input']
    optional_inputs = []

    def __init__(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        super().__init__(
            opengl_context = opengl_context
        )

    def _instantiate_cpu_and_gpu_blenders(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        """
        *For internal use only*

        Instantiate the CPU and GPU blenders and return 
        them in that order.

        This method has been created to be used in both the
        '__init__' and the '__reinit__' methods.

        This method must be implemented by each class.
        """
        from yta_editor_nodes_cpu.blender.alpha import AlphaBlenderCPU
        from yta_editor_nodes_gpu.blender.alpha import AlphaBlenderGPU

        node_cpu = AlphaBlenderCPU()
        node_gpu = AlphaBlenderGPU(
            opengl_context = opengl_context
        )

        return node_cpu, node_gpu

    def blend(
        self,
        inputs: dict[str, Union['np.ndarray', 'moderngl.Texture']],
        output_size: Union[tuple[int, int], None] = None,
        do_use_gpu: bool = True,
        mix_weight: float = 1.0,
        blend_strength: float = 1.0
    # TODO: What about the output type (?)
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the provided 'base_input' and 'overlay_input'
        with GPU or CPU according to the internal flag.

        The `mix_weight` will determine how much of the result
        we keep against the base, where a `mix_weight == 1.0`
        means that we want the result at 100%, but a
        `mix_weight == 0.4` means a 40% of the result and a 60%
        of the base.
        """
        return self.process(
            inputs = inputs,
            output_size = output_size,
            do_use_gpu = do_use_gpu,
            mix_weight = mix_weight,
            blend_strength = blend_strength
        )
    
    # TODO: This method has to be reviewed in the general
    # class to be able to receive array parameters and
    # send them to the individual process as single parameter
    def blend_multiple_inputs(
        self,
        inputs: list[Union['np.ndarray', 'moderngl.Texture']],
        output_size: Union[tuple[int, int], None] = None,
        do_use_gpu: bool = True,
        mix_weights: Union[list[float], float] = 1.0,
        blend_strengths: Union[list[float], float] = 1.0,
    # TODO: What about the output type (?)
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the provided 'inputs' with GPU or CPU
        according to the internal flag.
        """
        return self.blend_multiple_inputs(
            inputs = inputs,
            output_size = output_size,
            do_use_gpu = do_use_gpu,
            mix_weights = mix_weights,
            blend_strengths = blend_strengths
        )


from yta_editor_nodes.processor.video.base import _VideoNodeProcessor
from typing import Union


class WavingFrameVideoNodeProcessor(_VideoNodeProcessor):
    """
    The frame but as if it is moving like a wave.

    Mandatory parameters
    - `base_input`

    Optional parameters
    - `None`
    """

    mandatory_inputs = ['base_input']
    optional_inputs = []

    def __init__(
        self,
        opengl_context: 'moderngl.Context'
    ):
        super().__init__(
            opengl_context = opengl_context
        )

    def _instantiate_cpu_and_gpu_processors(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        """
        *For internal use only*

        Instantiate the CPU and GPU procesors and return 
        them in that order.

        This method must be implemented by each class.
        """
        from yta_editor_nodes_cpu.processor.video.waving_frame import WavingFrameVideoNodeProcessorCPU
        from yta_editor_nodes_gpu.processor.video.waving_frame import WavingFrameVideoNodeProcessorGPU

        node_cpu = WavingFrameVideoNodeProcessorCPU()
        node_gpu = WavingFrameVideoNodeProcessorGPU(
            opengl_context = opengl_context
        )

        return node_cpu, node_gpu
    
    def process(
        self,
        inputs: dict[str, Union['np.ndarray', 'moderngl.Texture']],
        output_size: Union[tuple[int, int], None],
        do_use_gpu: bool = True,
        t: float = 0.0,
        amplitude: float = 0.05,
        frequency: float = 10.0,
        speed: float = 2.0,
        do_use_transparent_pixels: bool = False
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the provided `inputs` with GPU or CPU 
        according to the internal flag.
        """
        return super().process(
            inputs = inputs,
            output_size = output_size,
            do_use_gpu = do_use_gpu,
            t = t,
            amplitude = amplitude,
            frequency = frequency,
            speed = speed,
            do_use_transparent_pixels = do_use_transparent_pixels
        )


"""
The nodes that are able to make simple processing.
"""
from yta_editor_nodes.processor.black_and_white import BlackAndWhiteNodeProcessor
from yta_editor_nodes.processor.brightness import BrightnessNodeProcessor
from yta_editor_nodes.processor.color_contrast import ColorContrastNodeProcessor
from yta_editor_nodes.processor.selection_mask import SelectionMaskNodeProcessor


__all__ = [
    'BlackAndWhiteNodeProcessor',
    'BrightnessNodeProcessor',
    'ColorContrastNodeProcessor',
    'SelectionMaskNodeProcessor'
]


from yta_editor_nodes.processor.video.transitions.base import _TransitionProcessor
from typing import Union


class CircleClosingTransitionProcessor(_TransitionProcessor):
    """
    A transition between the frames of 2 videos in
    which the frames are mixed by generating a circle
    that decreases its size from the whole screen
    until disappearing.

    Class that is capable of doing some processing on
    an input by using the GPU or the CPU.
    """

    def __init__(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        super().__init__(
            opengl_context = opengl_context
        )

    def _instantiate_cpu_and_gpu_processors(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        """
        *For internal use only*

        Instantiate the CPU and GPU processors and return 
        them in that order.
        """
        from yta_editor_nodes_gpu.processor.video.transitions.circle_closing import CircleClosingTransitionNodeProcessorGPU

        node_cpu = None
        node_gpu = CircleClosingTransitionNodeProcessorGPU(
            opengl_context = opengl_context
        )

        return node_cpu, node_gpu
    
    def process(
        self,
        inputs: dict[str, Union['moderngl.Texture', 'np.ndarray']],
        progress: float,
        output_size: Union[tuple[int, int], None],
        do_use_gpu: bool = True,
        border_smoothness: float = 0.02
    ) -> Union['moderngl.Texture', 'np.ndarray']:
        return super().process(
            inputs = inputs,
            progress = progress,
            output_size = output_size,
            do_use_gpu = do_use_gpu,
            border_smoothness = border_smoothness
        )